"""
Rough outline of how this module finds times of phenomena:

    * Split the time interval evenly into partitions. The partitions 
    should be small enough that no partitions contain multiple target values.
    * Evaluate the objective function at the partition edges to determine which 
    partitions contain the points of interest.
    * Pass the partitions that contain points of interest to the vectorized 
    optimization routines.
    
Some other things to be aware of are:
    
    * The variable ``f`` is the plain objective function that returns numbers 
    directly from skyfield.
    * The variable ``g`` is ``f`` transformed such that either all the target 
    values appear to the secant method as roots, or such that all of the 
    extremes appear to Brent's method as minima.
    
"""
from skyfield.api import load, Topos, EarthSatellite, Angle, Star
from skyfield.constants import tau
from skyfield.vectorlib import VectorSum
from skyfield.elementslib import osculating_elements_of
from optimizelib import secant, brent_min
from numpy import (degrees, arcsin, isfinite, hstack, nan, empty, linspace, 
                   ceil, sign, nonzero, zeros, empty_like, array)
from functools import partial

__all__ = ['meridian_transits', 'culminations', 'risings_settings', 
           'twilights', 'seasons', 'moon_phases']

ts = load.timescale()
    
#%% functions for finding/ isolating partitions
            
def _find_value(f, values, partition_edges, slope='any'):
    """ Evaluates ``f`` at ``partition_edges`` and returns the left and right 
    edges of those partitions that contain ``values``.
    
    Arguments
    ---------
    f : function
        Objective function. Must accept and return ndarrays.
    values : list
        List of values of ``f`` that are to be found
    partition_edges : ndarray
        Array of partition edges
    slope : str
        'positive' to find ``values`` when the slope is positive
        'negative' to find ``values`` when the slope is negative
        'any' to find ``values`` at any slope
        
    Returns
    -------
    jd0 : ndarray
        Left edges of those partitions found to contain ``values``
    jd1 : ndarray
        Right edges of those partitions found to contain ``values``
    targets : ndarray
        target from ``values`` that each partition is found to contain
    f0 : ndarray
        ``f(jd0)``
    f1 : ndarray
        ``f(jd1)``
    is_positive : ndarray, dtype=bool
        whether or not the slope of ``f`` in the partition is positive
    """
    f_values = f(partition_edges)             
                
    targets = empty(len(partition_edges)-1)
    targets.fill(nan)
    
    for value in values:
        g_values = (f_values - value + 180)%360 - 180
        if slope=='positive':
            found = (sign(g_values[:-1])==-1) * (sign(g_values[1:])==1)
        elif slope=='negative':
            found = (sign(g_values[:-1])==1) * (sign(g_values[1:])==-1)
        elif slope=='any':
            found = sign(g_values[:-1]) != sign(g_values[1:])
        
        if (isfinite(targets) * found).any():
            raise ValueError('Multiple target values found in the same partition. Make the partitions smaller.')
            
        targets[found] = value
        
    indices = nonzero(isfinite(targets))[0]
    jd0 = partition_edges[indices]
    jd1 = partition_edges[indices+1]
    f0 = f_values[indices]
    f1 = f_values[indices+1]
    
    is_positive = ((sign(g_values[:-1])==-1) * (sign(g_values[1:])==1))[indices]
    
    return jd0, jd1, targets[indices], f0, f1, is_positive


def _find_extremes(f, partition_edges, find='min'):
    """ Evaluates the derivative of ``f`` at ``partition_edges`` and returns 
    the left and right edges of the partitions that contain extrema.
    
    Arguments
    ---------
    f : function
        Objective function. Must accept and return ndarrays.
    partition_edges : ndarray
        Array of partition edges
    find : str
        'min' to find minima
        'max' to find maxima
        'any' to find all extrema
        
    Returns
    -------
    jd0 : ndarray
        Left edges of those partitions found to contain ``values``
    jd1 : ndarray
        Right edges of those partitions found to contain ``values``
    targets : ndarray
        target from ``values`` that each partition is found to contain
    f0 : ndarray
        ``f(jd0)``
    f1 : ndarray
        ``f(jd1)``
    minimum : ndarray, dtype=bool
        whether or not each partition contains a minima (False means the 
        partition contains a maxima)
    """  
    # evaluate the derivative using forward difference method. Combine both 
    # sides of the approximation so only one function call is necessary.
    step_size = 1e-6
    combined_array = hstack([partition_edges, partition_edges+step_size])
    combined_results = f(combined_array)
    left = combined_results[:len(partition_edges)]
    right = combined_results[len(partition_edges):]
    f_dot = (right - left)/step_size
    
    has_extreme = zeros(len(partition_edges)-1)
    
    if find == 'min' or find == 'any':
        has_min = (sign(f_dot[:-1])==-1) * (sign(f_dot[1:])==1)
        has_extreme[has_min] = -1
    
    if find == 'max' or find == 'any':
        has_max = (sign(f_dot[:-1])==1) * (sign(f_dot[1:])==-1)
        if find == 'any' and (has_min * has_max).any():
            raise ValueError('Multiple extremes found in the same partition. Make the partitions smaller.')
        has_extreme[has_max] = 1
        
    indices = nonzero(has_extreme)[0]
    jd0 = partition_edges[indices]
    jd1 = partition_edges[indices+1]
    minimum = has_extreme[indices]==-1
    f0 = left[indices]
    f1 = left[indices+1]
    
    return jd0, jd1, minimum, f0, f1


def _divide_evenly(start, end, max_width):
    """ Evenly divides the interval between ``start`` and ``end`` into 
    intervals that are at most ``max_width`` wide.
    
    Arguments
    ---------
    start : float
        Start of the interval
    end : float
        End of the interval
    max_width : float
        Maximum width of the divisions
        
    Returns
    -------
    divisions : ndarray
        Resulting array
    """
    num_partitions = int(ceil((end - start)/max_width))
    return linspace(start, end, num_partitions+1)

def _sidereal_period(secondary, primary):
    elements = osculating_elements_of((secondary-primary).at(ts.utc(2017)))
    return elements.period_in_days


#%% Objective Functions

def _ra(observer, body, t):
    """Returns the right ascension of 'body' in degrees at terrestrial time 't' 
    when seen from 'observer'.
    """
    return observer.at(ts.tt(jd=t)).observe(body).apparent().radec(epoch='date')[0]._degrees


def _ecliptic_lon(observer, body, t):
    """Returns the ecliptic latitude of body in degrees at terrestrial time t.
    """
    return observer.at(ts.tt(jd=t)).observe(body).apparent().ecliptic_latlon(epoch='date')[1].degrees


def _ecliptic_lon_diff(observer, body1, body2, t):
    """Returns the ecliptic longitudes of body1 minus that of body2
    in degrees at terrestrial time t.
    """
    diff = _ecliptic_lon(observer, body1, t) - _ecliptic_lon(observer, body2, t)
    return (diff + 180)%360 - 180


def _local_sidereal(observer, t):
    """Returns observer's local apparent sidereal time at t in degrees"""
    return ts.tt(jd=t).gast*15 + observer.vector_functions[-1].longitude.degrees


def _lha(observer, body, t):
    """Returns the local hour angle of `body` in degrees when seen from 
    `observer` at terrestrial time `t`.
    """
    return _local_sidereal(observer, t) - _ra(observer, body, t)


def _alt(observer, body, t):
    """Returns the altitude of `body` in degrees when seen from `observer` at 
    terrestrial time `t`.
    """
    return observer.at(ts.tt(jd=t)).observe(body).apparent().altaz()[0].degrees


def _moon_ul_alt(observer, moon, t):
    """Returns the altitude of the moon's upper limb in degrees when seen from
    observer at terrestrial time t.
    
    Calculates the moon's apparent semidiameter from the location of 
    ``observer``
    """
    moon_distance = observer.at(ts.tt(jd=t)).observe(moon).apparent().distance().au
    moon_radius = 1.161781e-5 #au
    moon_sd = degrees(arcsin(moon_radius/moon_distance))
    return _alt(observer, moon, t) + moon_sd


def _satellite_alt(observer, satellite, t):
    """Returns the altitude of `satellite` in degrees when seen from `observer` 
    at terrestrial time `t`.
    """
    return (satellite - observer).at(ts.tt(jd=t)).altaz()[0].degrees

def _linear_dist(observer, body, t):
    """Returns the geometric distance between observer and body in au at 
    terrestrial time t.
    """
    return (body - observer).at(ts.tt(jd=t)).distance().au


def _ecliptic_lat(observer, body, t):
    """Returns geometric ecliptic latitude in degrees at terrestrial time `t`.
    """
    return (body - observer).at(ts.tt(jd=t)).ecliptic_latlon()[0].degrees


#%% Topocentric Phenomena, pg. 482 in Explanatory Supplement (1992)

def meridian_transits(observer, body, t0, t1):
    """Calculates times of upper and lower transits of the local meridian.
    
    This function searches between ``t0`` and ``t1`` for times when ``body``'s 
    local hour angle is 0h for an upper transit or 12h for a lower transit.

    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> sun = ephem['sun']
    >>> earth = ephem['earth']
    >>> greenwich = earth + Topos('51.5 N', '0 W')
    >>> t0 = ts.utc(2017, 1, 1)
    >>> t1 = ts.utc(2017, 1, 8)
    >>> times, hour_angles = meridian_transits(greenwich, mars, t0, t1)
    >>>
    >>> upper_transits = times[hour_angles.hours==0]
    >>> lower_transits = times[hour_angles.hours==12]
    
    Parameters
    ----------
    observer : VectorSum
        VectorSum of earth + Topos
    body : Segment or VectorSum
        Vector representing the object whose transits are being found.
    t0 : Time
        Time object of length 1 representing the start of the search interval
    t1 : Time
        Time object of length 1 representing the end of the search interval
        
    Returns
    -------
    times : Time
        Times of transits
    hour_angles : Angle
        Local Hour Angle of ``body`` at ``times``
    """ 
    if isinstance(observer, Topos):
        raise ValueError('`observer` should be a VectorSum of earth + Topos.')
    
    if isinstance(body, EarthSatellite) or (isinstance(body, VectorSum) and isinstance(body.vector_functions[-1], EarthSatellite)):
        raise ValueError("meridian_transits doesn't support EarthSatellites.")
    
    f = partial(_lha, observer, body)    
    partition_edges = _divide_evenly(t0.tt, t1.tt, .2)
    
    jd0, jd1, targets, f0, f1, _ = _find_value(f, [0, 180], partition_edges, slope='positive')
    
    times = secant(f, jd0, jd1, targets, f0, f1)
    
    return ts.tt(jd=times), Angle(degrees=targets, preference='hours')


def culminations(observer, body, t0, t1):
    """Calculates times of upper and lower culminations.
    
    This function searches between ``t0`` and ``t1`` for times when `body`'s 
    altitude reaches a local maximum or minimum.

    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> sun = ephem['sun']
    >>> earth = planets['earth']
    >>> greenwich = earth + Topos('51.5 N', '0 W')
    >>> t0 = ts.utc(2017, 1, 1)
    >>> t1 = ts.utc(2017, 1, 8)
    >>> times, kinds = culminations(greenwich, sun, t0, t1)
    >>>
    >>> upper_culminations = times[kinds=='upper']
    >>> lower_culminations = times[kinds=='lower']
    
    Parameters
    ----------
    observer : VectorSum or Topos
        VectorSum of earth + Topos. If ``body`` is an EarthSatellite, 
        ``observer`` can be either a VectorSum or a plain Topos object.
    body : Segment, VectorSum, or EarthSatellite
        Vector representing the object whose culminations are being found. For 
        EarthSatellites use a plain Earthsatellite and not a VectorSum.
    t0 : Time
        Time object of length 1 representing the start of the search interval
    t1 : Time
        Time object of length 1 representing the end of the search interval
        
    Returns
    -------
    times : Time
        Times of altitude maximums or minimums
    kinds : ndarray, dtype=str
        array containing 'upper' for upper culminations, or 'lower' for lower 
        culminations
    """    
    if isinstance(body, EarthSatellite) and isinstance(observer, VectorSum):
        observer = observer.vector_functions[-1]
        
    if not isinstance(body, EarthSatellite) and not isinstance(observer, VectorSum):
        raise ValueError('Unless ``body`` is an EarthSatellite, ``observer``'
                         'must be a VectorSum of Earth + Topos')
        
    if isinstance(body, VectorSum) and isinstance(body.vector_functions[-1], EarthSatellite):
        raise ValueError('`body` should be a plain EarthSatellite, not a VectorSum')
    
    if isinstance(body, EarthSatellite):
        f = partial(_satellite_alt, observer, body)
        period = tau/body.model.no/60/24 # days/orbit
        max_width = period * .2
    else:
        f = partial(_alt, observer, body)
        max_width = .2
    
    partition_edges = _divide_evenly(t0.tt, t1.tt, max_width)

    jd0, jd1, minimum, f0, f1 = _find_extremes(f, partition_edges, find='any')

    if len(jd0)!=0:
        times = brent_min(f, jd0, jd1, minimum, f0, f1, tol=1e-15)
    else:
        times = array([])

    kinds = empty_like(minimum, dtype='U5')
    kinds[minimum] = 'lower'
    kinds[~minimum] = 'upper'

    return ts.tt(jd=times), kinds
    

def risings_settings(observer, body, t0, t1):
    """Calculates times when an object rises and sets.
    
    This function searches between ``t0`` and ``t1`` for times when `body`'s 
    altitude (uncorrected for refraction) is -34 arcminutes. The sun's and 
    moon's upper limb is used rather than their center.

    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> sun = ephem['sun']
    >>> earth = ephem['earth']
    >>> greenwich = earth + Topos('51.5 N', '0 W')
    >>> t0 = ts.utc(2017, 1, 1)
    >>> t1 = ts.utc(2017, 1, 8)
    >>> times, kinds = risings_settings(greenwich, sun, t0, t1)
    >>>
    >>> risings = times[kinds=='rise']
    >>> settings = times[kinds=='set']
    
    Parameters
    ----------
    observer : VectorSum or Topos
        VectorSum of earth + Topos. If ``body`` is an EarthSatellite, 
        ``observer`` can be either a VectorSum or a plain Topos object.
    body : Segment, VectorSum, or EarthSatellite
        Vector representing the object whose rise/set times are being found. 
        For EarthSatellite objects use a plain EarthSatellite and not a 
        VectorSum.
    t0 : Time
        Time object of length 1 representing the start of the search interval
    t1 : Time
        Time object of length 1 representing the end of the search interval
        
    Returns
    -------
    times : Time
        Times that `body` rises or sets
    kinds : ndarray, dtype=str
        array containing 'rise' for risings, or 'set' for settings
    """
    if isinstance(body, EarthSatellite) and isinstance(observer, VectorSum):
        observer = observer.vector_functions[-1]
    
    if not isinstance(body, EarthSatellite) and not isinstance(observer, VectorSum):
        raise ValueError('Unless ``body`` is an EarthSatellite, ``observer``'
                         'must be a VectorSum of Earth + Topos')
        
    if isinstance(body, VectorSum) and isinstance(body.vector_functions[-1], EarthSatellite):
        raise ValueError('`body` should be a plain EarthSatellite, not a VectorSum')
    
    if isinstance(body, EarthSatellite):
        f = partial(_satellite_alt, observer, body)
        value = [-34/60]
    elif isinstance(body, Star):
        f = partial(_alt, observer, body)
        value = [0]
    elif body.target == 10: # sun
        f = partial(_alt, observer, body)
        value = [-50/60]
    elif body.target == 301: # moon
        f = partial(_moon_ul_alt, observer, body)
        value = [-34/60]
    else:
        f = partial(_alt, observer, body)
        value = [0]
    
    body_culminations = culminations(observer, body, t0, t1)[0].tt
    partition_edges = hstack([t0.tt, body_culminations, t1.tt])    
    
    jd0, jd1, targets, f0, f1, is_positive = _find_value(f, value, partition_edges)
    
    times = secant(f, jd0, jd1, targets, f0, f1, tol=1e-15)
    
    kinds = empty_like(is_positive, dtype='U4')
    kinds[is_positive] = 'rise'
    kinds[~is_positive] = 'set'

    return ts.tt(jd=times), kinds
    

def twilights(observer, sun, t0, t1, kind='civil'):
    """Calculates times when twilight starts or ends in the morning or evening.
   
    This function searches between ``t0`` and ``t1`` for times when the sun's 
    altitude is -6, -12, or -18 degrees for civil, nautical, or astronomical 
    twilights, respectively.

    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> earth = ephem['earth']
    >>> greenwich = earth + Topos('51.5 N', '0 W')
    >>> t0 = ts.utc(2017, 1, 1)
    >>> t1 = ts.utc(2017, 1, 8)
    >>> times, am_pm = twilights(greenwich, sun, t0, t1, kind='nautical')
    >>>
    >>> am_twilights = times[am_pm=='am']
    >>> pm_twilights = times[am_pm=='pm']
    
    Parameters
    ----------
    observer : VectorSum
        VectorSum of earth + Topos
    sun : Segment
        Segment representing the sun
    t0 : Time
        Time object of length 1 representing the start of the search interval
    t1 : Time
        Time object of length 1 representing the end of the search interval
    kind : str
        ``'civil'`` for civil twilight
        ``'nautical'`` for nautical twilight
        ``'astronomical'`` for astronomical twilight
        
    Returns
    -------
    times : Time
        Times that twilight starts in the morning and/ or ends in the evening
    am_pm : ndarray, dtype=str
        array containing 'am' for morning twilight, or 'pm' for evening twilight
        """
    if not isinstance(observer.vector_functions[-1], Topos):
        raise ValueError('`observer` should be a VectorSum of earth + Topos.')
        
    sun_culminations = culminations(observer, sun, t0, t1)[0].tt
    partition_edges = hstack([t0.tt, sun_culminations, t1.tt])
    
    f = partial(_alt, observer, sun)
    if kind == 'civil':
        value = [-6]
    elif kind == 'nautical':
        value = [-12]
    elif kind == 'astronomical':
        value = [-18]
    else:
        raise ValueError("kind must be 'civil', 'nautical', or 'astronomical'")

    jd0, jd1, targets, f0, f1, is_positive = _find_value(f, value, partition_edges)

    times = secant(f, jd0, jd1, targets, f0, f1)

    am_pm = empty_like(is_positive, dtype='U2')
    am_pm[is_positive] = 'am'
    am_pm[~is_positive] = 'pm'

    return ts.tt(jd=times), am_pm
    
#%% Geocentric Phenomena, pg. 478 in Explanatory Supplement (1992)

def seasons(earth, t0, t1):
    """Calculates times of equinoxes and solstices.
    
    This function searches between ``t0`` and ``t1`` for times when the sun's 
    ecliptic longitude is:
        
        * 0 degrees for march equinoxes
        * 90 degrees for june solstices
        * 180 degrees for september equinoxes
        * 270 degrees for december solstices
    
    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> earth = ephem['earth']
    >>> t0 = ts.utc(2017)
    >>> t1 = ts.utc(2018)
    >>> times, lons = seasons(earth, t0, t1)
    >>>
    >>> march_equinoxes = times[lons.degrees==0]
    >>> june_solstices = times[lons.degrees==90]
    >>> sept_equinoxes = times[lons.degrees==180]
    >>> dec_solstices = times[lons.degrees==270]
    
    Parameters
    ----------
    earth : Segment
        Vector representing earth
    t0 : Time
        Time object of length 1 representing the start of the search interval
    t1 : Time
        Time object of length 1 representing the end of the search interval
        
    Returns
    -------
    times : Time
        Times of solstices
    longitudes : Angle
        sun's ecliptic longitude at ``times``
    """
    sun = earth.ephemeris['sun']

    f = partial(_ecliptic_lon, earth, sun)

    partition_edges = _divide_evenly(t0.tt, t1.tt, 365*.2)

    jd0, jd1, targets, f0, f1, _ = _find_value(f, [0, 90, 180, 270], partition_edges, slope='positive')
        
    times = secant(f, jd0, jd1, targets, f0, f1)

    return ts.tt(jd=times), Angle(degrees=targets)


def moon_phases(moon, t0, t1):
    """Calculates times of the phases of the moon.
    
    This function searches between ``t0`` and ``t1`` for times when the moon's 
    geocentric ecliptic longitude minus that of the sun is:
        
        * 0 degrees for new moon
        * 90 degrees for first quarter
        * 180 degrees for full moon
        * 270 degrees for last quarter 
    
    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> moon = ephem['moon']
    >>> t0 = ts.utc(2017, 1)
    >>> t1 = ts.utc(2017, 2)
    >>> times, lon_diffs = moon_phases(moon, t0, t1)
    >>>
    >>> new_moons = times[lon_diffs.degrees==0]
    >>> first_quarters = times[lon_diffs.degrees==90]
    >>> full_moons = times[lon_diffs.degrees==180]
    >>> last_quarters = times[lon_diffs.degrees==270]
    
    Parameters
    ----------
    moon : Segment
        Vector representing the moon
    t0 : Time
        Time object of length 1 representing the start of the search interval
    t1 : Time
        Time object of length 1 representing the end of the search interval
        
    Returns
    -------
    times : Time
        Times of moon quarters
    longitude_diffs: Angle
        moon's ecliptic longitude - sun's ecliptic longitude at ``times``
        
    """ 
    earth = moon.ephemeris['earth']
    sun = moon.ephemeris['sun']
    
    f = partial(_ecliptic_lon_diff, earth, moon, sun)
    
    partition_edges = _divide_evenly(t0.tt, t1.tt, 29*.2)

    jd0, jd1, targets, f0, f1, _ = _find_value(f, [0, 90, 180, 270], partition_edges, slope='positive')
        
    times = secant(f, jd0, jd1, targets, f0, f1)

    return ts.tt(jd=times), Angle(degrees=targets)

#%% Heliocentric Phenomena, pg. 478 in Explanatory Supplement (1992)
    
def apsides(secondary, primary, t0, t1):
    """Calculates times of periapsides and apoapsides.
    
    This function searches between ``t0`` and ``t1`` for times when 
    `secondary`'s distance from `primary` is a maximum or minimum.

    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> earth = ephem['earth']
    >>> sun = ephem['sun']
    >>> t0 = ts.utc(2017)
    >>> t1 = ts.utc(2018)
    >>> apsides(earth, sun, t0, t1)
    
    Parameters
    ----------
    secondary : Segment or VectorSum
        Vector representing the object whose apsides are being found
    primary : Segment or VectorSum
        Vector representing the primary of the object whose apsides are being 
        found
    t0 : Time
        The start of the time range to search
    t1 : Time
        The end of the time range to search
        
    Returns
    -------
    times : Time
        Times of apsides
    kinds : ndarray, dtype=str
        array containing 'apo' for apoapsides, or 'peri' for periapsides
    """
    
    f = partial(_linear_dist, secondary, primary)
    
    period = _sidereal_period(secondary, primary)
    partition_edges = _divide_evenly(t0.tt, t1.tt, period*.2)
        
    jd0, jd1, minimum, f0, f1 = _find_extremes(f, partition_edges, find='any')

    if len(jd0)!=0:
        times = brent_min(f, jd0, jd1, minimum, f0, f1, tol=1e-13)
    else:
        times = array([])

    kinds = empty_like(minimum, dtype='U5')
    kinds[minimum] = 'peri'
    kinds[~minimum] = 'apo'

    return ts.tt(jd=times), kinds


def nodes(secondary, primary, t0, t1):
    """Calculates times of ascending and descending nodes.

    This function searches between ``t0`` and ``t1``
    for times when `secondary`'s ecliptic latitude is 0 degrees.
    
    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> sun = ephem['sun']
    >>> mars = planets['mars']
    >>> t0 = ts.utc(2017)
    >>> t1 = ts.utc(2018)
    >>> times, kinds = nodes(mars, sun, t0, t1)
    
    Parameters
    ----------
    secondary : Segment or VectorSum
        Vector representing the object whose nodes are being found
    primary : Segment or VectorSum
        Vector representing the object orbited by `secondary`
    t0 : Time
        The start of the time range to search
    t1 : Time
        The end of the time range to search

    Returns
    -------
    times : Time
        Times of nodes
    kinds : ndarray, dtype=str
        array containing 'ascending' for apoapsides, or 'descending' for periapsides
    """
    
    period = _sidereal_period(secondary, primary)
    partition_edges = _divide_evenly(t0.tt, t1.tt, period*.2)   
    
    f = partial(_ecliptic_lat, secondary, primary)
    
    jd0, jd1, targets, f0, f1, is_positive = _find_value(f, [0], partition_edges)
    
    times = secant(f, jd0, jd1, targets, f0, f1, tol=1e-10)
    
    kinds = empty_like(is_positive, dtype='U10')
    kinds[is_positive] = 'ascending'
    kinds[~is_positive] = 'descending'

    return ts.tt(jd=times), kinds


def max_ecliptic_latitudes(secondary, primary, t0, t1):
    """Calculates data about extreme max and min ecliptic latitudes.

    This function searches between ``t0`` and ``t1``
    for times when `secondary`'s ecliptic latitude centered at ``primary`` is at a 
    local maximum, e.g., if ``primary`` is the sun, the ecliptic latitudes will 
    be heliocentric.

    Example
    -------
    >>> ephem = load('de430t.bsp')
    >>> venus = ephem['venus']
    >>> sun = ephem['sun']
    >>> t0 = ts.utc(2017)
    >>> t1 = ts.utc(2020)
    >>> times, lats = extreme_ecliptic_latitudes(venus, sun, t0, t1)
    >>> north_times = times[lats>0]
    >>> south_times = times[lats<0]
    
    Parameters
    ----------
    secondary : Segment or VectorSum
        Vector representing the object whose maximum ecliptic latitudes are 
        being found
    primary : Segment or VectorSum
        Vector representing the object orbited by `secondary`
    t0 : Time
        The start of the time range to search
    t1 : Time
        The end of the time range to search

    Returns
    -------
    times : Time
        Times of extreme ecliptic latitudes
    lats : Angle
        Ecliptic latitudes that correspond to `times`
    """
    
    period = _sidereal_period(secondary, primary)
    partition_edges = _divide_evenly(t0.tt, t1.tt, period*.2)   
    
    f = partial(_ecliptic_lat, secondary, primary)

    jd0, jd1, minimum, f0, f1 = _find_extremes(f, partition_edges, find='any')

    if len(jd0)!=0:
        times = brent_min(f, jd0, jd1, minimum, f0, f1, tol=1e-15)
    else:
        times = array([])
        
    # TODO: Should brent_min retur f(x) to avoid this function call?
    lats = f(times)

    return ts.tt(jd=times), Angle(degrees=lats)



from numpy import array
from time import time

class A(object):
    __getitem__ = array
A = A()

t0 = time()
x = array((1, 2, 3))
tt = time() - t0
print('Old way:', tt)

t0 = time()
x = A[1, 2, 3]
tt = time() - t0
print('A[] way:', tt)





from skyfield import api
from skyfield.data.tycho2 import load_dataframe

def test_dataframe():
    with api.load.open('tyc_main_head.dat') as f:
        df = load_dataframe(f)
    star = api.Star.from_dataframe(df)
    assert repr(star) == 'Star(ra shape=10, dec shape=10, ra_mas_per_year shape=10, dec_mas_per_year shape=10, parallax_mas shape=10, epoch shape=10)'



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.collections import LineCollection

from skyfield.api import Star, load
from skyfield.constants import GM_SUN_Pitjeva_2005_km3_s2 as GM_SUN
from skyfield.data import hipparcos, mpc, stellarium
from skyfield.projections import build_stereographic_projection

# The comet is plotted on several dates `t_comet`.  But the stars only
# need to be drawn once, so we take the middle comet date as the single
# time `t` we use for everything else.

ts = load.timescale()
t_comet = ts.utc(2020, 7, range(17, 27))
t = t_comet[len(t_comet) // 2]  # middle date

# An ephemeris from the JPL provides Sun and Earth positions.

eph = load('de421.bsp')
sun = eph['sun']
earth = eph['earth']

# The Minor Planet Center data file provides the comet orbit.

with load.open(mpc.COMET_URL) as f:
    comets = mpc.load_comets_dataframe(f)

comets = (comets.sort_values('reference')
          .groupby('designation', as_index=False).last()
          .set_index('designation', drop=False))

row = comets.loc['C/2020 F3 (NEOWISE)']
comet = sun + mpc.comet_orbit(row, ts, GM_SUN)

# The Hipparcos mission provides our star catalog.

with load.open(hipparcos.URL) as f:
    stars = hipparcos.load_dataframe(f)

# And the constellation outlines come from Stellarium.  We make a list
# of the stars at which each edge stars, and the star at which each edge
# ends.

url = ('https://raw.githubusercontent.com/Stellarium/stellarium/master'
       '/skycultures/modern_st/constellationship.fab')

with load.open(url) as f:
    constellations = stellarium.parse_constellations(f)

edges = [edge for name, edges in constellations for edge in edges]
edges_star1 = [star1 for star1, star2 in edges]
edges_star2 = [star2 for star1, star2 in edges]

# We will center the chart on the comet's middle position.

center = earth.at(t).observe(comet)
projection = build_stereographic_projection(center)
field_of_view_degrees = 45.0
limiting_magnitude = 7.0

# Now that we have constructed our projection, compute the x and y
# coordinates that each star and the comet will have on the plot.

star_positions = earth.at(t).observe(Star.from_dataframe(stars))
stars['x'], stars['y'] = projection(star_positions)

comet_x, comet_y = projection(earth.at(t_comet).observe(comet))

# Create a True/False mask marking the stars bright enough to be
# included in our plot.  And go ahead and compute how large their
# markers will be on the plot.

bright_stars = (stars.magnitude <= limiting_magnitude)
magnitude = stars['magnitude'][bright_stars]
marker_size = (0.5 + limiting_magnitude - magnitude) ** 2.0

# The constellation lines will each begin at the x,y of one star and end
# at the x,y of another.  We have to "rollaxis" the resulting coordinate
# array into the shape that matplotlib expects.

xy1 = stars[['x', 'y']].loc[edges_star1].values
xy2 = stars[['x', 'y']].loc[edges_star2].values
lines_xy = np.rollaxis(np.array([xy1, xy2]), 1)

# Time to build the figure!

fig, ax = plt.subplots(figsize=[9, 9])

# Draw the constellation lines.

ax.add_collection(LineCollection(lines_xy, colors='#00f2'))

# Draw the stars.

ax.scatter(stars['x'][bright_stars], stars['y'][bright_stars],
           s=marker_size, color='k')

# Draw the comet positions, and label them with dates.

comet_color = '#f00'
offset = 0.002

ax.plot(comet_x, comet_y, '+', c=comet_color, zorder=3)

for xi, yi, tstr in zip(comet_x, comet_y, t_comet.utc_strftime('%m/%d')):
    tstr = tstr.lstrip('0')
    text = ax.text(xi + offset, yi - offset, tstr, color=comet_color,
                   ha='left', va='top', fontsize=9, weight='bold', zorder=-1)
    text.set_alpha(0.5)

# Finally, title the plot and set some final parameters.

angle = np.pi - field_of_view_degrees / 360.0 * np.pi
limit = np.sin(angle) / (1.0 - np.cos(angle))

ax.set_xlim(-limit, limit)
ax.set_ylim(-limit, limit)
ax.xaxis.set_visible(False)
ax.yaxis.set_visible(False)
ax.set_aspect(1.0)
ax.set_title('Comet NEOWISE {} through {}'.format(
    t_comet[0].utc_strftime('%Y %B %d'),
    t_comet[-1].utc_strftime('%Y %B %d'),
))

# Save.

fig.savefig('neowise-finder-chart.png', bbox_inches='tight')



from skyfield.api import load, wgs84
from numpy import array, nan

from skyfield.functions import dots
from skyfield.geometry import line_and_ellipsoid_intersection

def vector_projection(a, b):
    return dots(a,b) / dots(b,b) * b

assert list(vector_projection(array([6.0,  6.0, 6.0]),
             array([2.0,  0.0,  0.0]))) == ([6.0,  0.0,  0.0])

assert list(vector_projection(array([6.0,  6.0, 6.0]),
             array([-3.0,  0.0,  0.0]))) == ([6.0, -0.0, -0.0])

assert list(vector_projection(array([6.0,  6.0, 0.0]),
             array([0.0,  7.0,  0.0]))) == ([0.0,  6.0,  0.0])

assert list(vector_projection(array([6.0,  0.0, 0.0]),
             array([0.0,  0.0,  9.0]))) == ([0.0,  0.0,  0.0])

# And now, start work with the real position.

def main():
    ts = load.timescale()
    eph = load('de421.bsp')
    sun = eph['sun']
    moon = eph['Moon']
    earth = eph['earth']

    #t = ts.utc(2017, 8, 21, 18, 30)
    t = ts.utc(2017, 8, 21, 16, range(0, 300, 15))

    e = earth.at(t)
    s = e.observe(sun).apparent()
    m = e.observe(moon).apparent()

    A = B = wgs84.radius.km  # x and y axes point to equator 6378.0
    C = wgs84.polar_radius.km

    print('Intersection, in ITRF coordinates:')
    shadow_itrf = line_and_ellipsoid_intersection(
        s.frame_xyz(itrs).km,
        m.frame_xyz(itrs).km - s.frame_xyz(itrs).km,
        array([A, B, C]),
    )
    print(shadow_itrf)

    # So where's that on Earth?
    from skyfield.positionlib import ICRF, Distance, Velocity
    position = ICRF.from_time_and_frame_vectors(
        t, itrs, Distance.km(shadow_itrf), Velocity(0),
    )
    position.center = 399
    g = wgs84.geographic_position_of(position)
    print(g)

    latitude = g.latitude.degrees
    longitude = g.longitude.degrees
    midpoint = len(latitude) // 2

    import matplotlib.pyplot as plt
    import cartopy.crs as ccrs

    o = ccrs.Orthographic(
        central_longitude=longitude[midpoint],
        central_latitude=latitude[midpoint],
    )

    fig = plt.figure(figsize=(3, 3))
    ax = plt.axes(projection=o)
    ax.coastlines(resolution='110m')
    ax.gridlines()
    ax.set_global()
    ax.plot(
        longitude, latitude,
        color='blue', linewidth=1, marker='+',
        transform=ccrs.Geodetic(),
    )

    fig.savefig('eclipse.png')

from skyfield.framelib import itrs

class A(object):
    __getitem__ = array
A = A()

nan3 = A[nan,nan,nan]

def test():
    f = line_and_ellipsoid_intersection
    assert list(f(A[-5,0,0], A[1,0,0], A[1,2,3])) == [-1,0,0]
    assert list(f(A[-5,2,0], A[1,0,0], A[1,2,3])) == [0,2,0]
    assert list(f(A[5,0,3], A[-1,0,0], A[1,2,3])) == [0,0,3]

    starts = A[[-5,0,0], [-5,2,0], [0,-5,0], [0,-5,4], [0,0,-5], [1,0,-5]]
    surface_points = A[[1,0,0], [0,2,0], [0,0,4], [-1,0,0], [0,-2,0], [0,0,-4]]
    radii = A[1,2,4]

    def check(starts, answers):
        for start, answer in zip(starts, answers):
            actual = f(start * 1.0, direction * 1.0, radii)
            print(start, direction, answer, '?=', actual)
            assert str(answer * 1.0) == str(actual)

    direction = A[1,0,0]
    answers = A[[-1,0,0], [0,2,0], nan3, nan3, nan3, nan3]
    check(starts, answers)
    check(surface_points, surface_points)

    direction = A[0,1,0]
    answers = A[nan3, nan3, [0,-2,0], [0,0,4], nan3, nan3]
    check(starts, answers)
    check(surface_points, surface_points)

    direction = A[0,0,1]
    answers = A[nan3, nan3, nan3, nan3, [0,0,-4], [1,0,0]]
    check(starts, answers)
    check(surface_points, surface_points)

    direction = A[-1,0,0]
    answers = A[nan3, nan3, nan3, nan3, nan3, nan3]
    check(starts, answers)
    check(surface_points, surface_points)

test()
main()

# Output:
# 'WGS84 latitude +36.2101 N longitude -85.9654 E elevation -3.1 m'
# Boom :)



"""Verdict: light-time delay from the Sun can be neglected in is_sunlit()!

For the Sun's position from Earth, does calling observe() really make
enough difference to justify the expense?  Here we take two approaches
to answering the question: we compare the difference every day over 40
years, and then we do a back-of-the-envelope estimate of how big we
might have expected the effect to be.  The two approaches agree!  The
maximum difference is around 10 mas.

What difference does that make for a satellite?  Let's take the ISS.
With its orbital period of 92 minutes, it sees the Earth swing in a full
circle around the sky in that amount of time.  That's 360/90 = 4 degrees
per minute (!) = 240 arcseconds per second.  At that speed, a difference
of 10 mas in the Sun's position would at most hasten or delay the moment
of sunrise for the ISS by 40 microseconds, which is far below the
accuracy of TLE position predictions and can thus be safely incurred.

"""
from skyfield import api
from skyfield.api import load

ts = load.timescale()
eph = load('de421.bsp')
sun = eph['sun']
earth = eph['earth']

t = ts.utc(2000, 1, range(40 * 365))
s1 = earth.at(t).observe(sun)
s2 = (sun - earth).at(t)
print('Milliarcseconds (mas) difference:', s1.separation_from(s2).mas().max())

print()
print('Does that make physical sense?')

# The Sun orbits around the Solar System barycenter which is usually
# inside the Sun's radius but occasionally a bit outside of it.  So we
# can very roughly imagine the Sun's orbit as its own circumference,
# give or take.
solar_radius_km = 696340
distance_sun_travels_in_one_orbit = solar_radius_km * api.tau

# It takes the Sun more than a decade to travel that path, as its orbit
# is roughly the opposite of Jupiter's (which takes 12 years to circle
# the Sun).  So it turns out that it travels a bit slowly.
sun_km_per_s = distance_sun_travels_in_one_orbit / 10 / 365.25 / 24 / 60 / 60
print('Sun km/s:', sun_km_per_s)

light_delay_seconds = s2[0].position.length().light_seconds()
print('Sample light delay from Sun to Earth (seconds):', light_delay_seconds)

print('How far does the Sun move in that time?')

travel_km = light_delay_seconds * sun_km_per_s
print('Sun km moved during that light travel time:', travel_km)

print('What angle does that many kilometers subtend from Earth?')

earth_sun_distance_km = 150e6
travel_angle = api.Angle(radians = travel_km / earth_sun_distance_km)
print('Angle traveled by sun in arcseconds:', travel_angle.arcseconds())
print('Angle traveled by sun in mas:', travel_angle.mas())

print()
print(__doc__.rstrip())



"""
The variable ``f`` is the plain objective function that returns numbers 
directly from skyfield.

The variable ``g`` is ``f`` transformed such that either all the target 
values appear to the secant method as roots, or such that all of the 
extremes appear to Brent's method as minima.
"""

import numpy

def secant(f, jd0, jd1, targets=0, f0=None, f1=None, tol=1e-10):
    """ Performs secant search to find target values in f.
    
    This is a vectorized version of the function newton from pyephem:
    https://github.com/brandon-rhodes/pyephem/blob/f96daf12d4f815be92e0caa52611b444517b9e0d/ephem/__init__.py#L91
    
    ``f`` is the plain objective function, and ``g`` is the objective function 
    translated vertically such that all of the target values appear to the 
    algorithm as roots, and renormalized to -180 to 180 degrees so that the 
    partitions won't contain any discontinuities.
    
    Arguments
    ---------
    f : function
        Objective function. Must accept and return ndarrays.
    jd0 : ndarray
        Left sides of partitions known to contain the ``targets``
    jd1 : ndarray
        Right sides of partitions known to contain the ``target``
    targets : float or ndarray
        Target values corresponding to the partitions defined by jd0 and jd1
    f0 : ndarray
        ``f(jd0)``. Providing this saves an extra function call
    f1 : ndarray
        ``f(jd1)``. Providing this saves an extra function call
    tol : float
        Tolerance used to determine when convergence is complete.
        
    Returns
    -------
    jd : ndarray
        the jd values at which ``f(jd) == targets``
    """
    if numpy.isscalar(targets):
        targets = numpy.ones_like(jd0) * targets
    
    def g(t, targets):
        return (f(t) - targets + 180)%360 - 180
    
    g0 = (f0 - targets + 180)%360 - 180 if f0 is not None else g(jd0, targets)
    g1 = (f1 - targets + 180)%360 - 180 if f1 is not None else g(jd1, targets)
    
    iters = 0
    max_iters = 50
    while iters < max_iters:   
        converged = (g1 == 0) + (abs(jd1 - jd0) < tol) + (g1 == g0)
        if converged.all():
            break
        inds = ~converged

        jd0[inds], jd1[inds] = jd1[inds], jd1[inds] + (jd1[inds] - jd0[inds]) / (g0[inds]/g1[inds] - 1)
        g0[inds], g1[inds] = g1[inds], g(jd1[inds], targets[inds])
        iters += 1
    return jd1


def _bracket(func, xa, xb, multiplier=None, f0=None, f1=None):
    """ This function is only meant to be used by brent_min
    """
    if multiplier is None:
        multiplier = numpy.ones_like(xa)
    
    _gold = 1.618034  # golden ratio: (1.0+sqrt(5.0))/2.0
    _verysmall_num = 1e-21
    grow_limit=110.0
    maxiter=1000
    failing = numpy.ones_like(xa, dtype=bool)
    
    def g(x, multiplier):
        return multiplier*func(x) + (multiplier==1)*360
    
    fa = g(xa, multiplier) if f0 is not None else multiplier*f0 + (multiplier==1)*360
    fb = g(xb, multiplier) if f1 is not None else multiplier*f1 + (multiplier==1)*360
    
    # Switch so fa > fb
    ind1 = fa<fb
    xa[ind1], xb[ind1] = xb[ind1], xa[ind1]
    fa[ind1], fb[ind1] = fb[ind1], fa[ind1]
    
    xc = xb + _gold * (xb - xa)
    fc = g(xc, multiplier)
    iter_ = 0
    
    failing[fc>=fb] = False
    
    while failing.any():
        tmp1 = (xb - xa) * (fb - fc)
        tmp2 = (xb - xc) * (fb - fa)
        val = tmp2 - tmp1
        
        denom = numpy.zeros_like(val)
        ind2 = numpy.abs(val) < _verysmall_num
        denom[ind2] = 2.0 * _verysmall_num
        denom[~ind2] = 2.0 * val
        
        w = xb - ((xb - xc) * tmp2 - (xb - xa) * tmp1) / denom
        wlim = xb + grow_limit * (xc - xb)
        if iter_ > maxiter:
            raise RuntimeError("Too many iterations.")
        iter_ += 1
        
        case1 = (w - xc) * (xb - w) > 0.0
        case2 = ~case1 * ((w - wlim)*(wlim - xc) >= 0.0)
        case3 = (~case1 + ~case2) * ((w - wlim)*(xc - w) > 0.0)
        case4 = ~case1 * ~case2 * ~case3
        
        fw = numpy.zeros_like(xa)
        
        fw[case1] = g(w[case1], multiplier[case1])
        
        ind1 = case1 * (fw < fc) * failing
        ind2 = case1 * ~ind1 * (fw > fb) * failing

        xa[ind1] = xb[ind1]
        xb[ind1] = w[ind1]
        fa[ind1] = fb[ind1]
        fb[ind1] = fw[ind1]
        failing[ind1] = False

        xc[ind2] = w[ind2]
        fc[ind2] = fw[ind2]
        failing[ind2] = False
            
        w[case1] = xc[case1] + _gold * (xc[case1] - xb[case1])
        fw[case1] = g(w[case1], multiplier[case1])

        w[case2] = wlim[case2]
        if case2.any():
            fw[case2] = g(w[case2], multiplier[case2])

        if case3.any():
            fw[case3] = g(w[case3], multiplier[case3])
        
        ind3 = case3 * (fw < fc) * failing
        xb[ind3] = xc[ind3]
        xc[ind3] = w[ind3]
        w[ind3] = xc[ind3] + _gold * (xc[ind3] - xb[ind3])
        fb[ind3] = fc[ind3]
        fc[ind3] = fw[ind3]
        
        if ind3.any():
            fw[ind3] = g(w[ind3], multiplier[ind3])

        w[case4] = xc[case4] + _gold * (xc[case4] - xb[case4])
        
        if case4.any():
            fw[case4] = g(w[case4], multiplier[case4])
        
        xa[failing] = xb[failing]
        xb[failing] = xc[failing]
        xc[failing] = w[failing]
        fa[failing] = fb[failing]
        fb[failing] = fc[failing]
        fc[failing] = fw[failing]
    return xa, xb, xc, fa, fb, fc


def brent_min(f, jd0, jd1, minimum=True, f0=None, f1=None, tol=1.48e-8):
    """ Vectorized version of Brent's method for finding maxima and minima.
    
    This is a vectorized version of scipy.optimize.optimize.Brent.optimize:
    https://github.com/scipy/scipy/blob/de9e0d6022be0319c1ba73c07a0946be46e02a24/scipy/optimize/optimize.py#L1896
    
    ``f`` is the plain objective function, and ``g`` is the objective function 
    optionally mirrored over the horizontal axis such that all of the extrema 
    appear to the algorithm as minima.
    
    Arguments
    ---------
    f : function
        Objective function. Must accept and return ndarrays.
    jd0 : ndarray
        Left sides of partitions known to contain maxima or minima
    jd1 : ndarray
        Right sides of partitions known to contain the target value
    minimum : bool or ndarray
        Whether or not the corresponding partitions contain minima. True means 
        the corresponding partition contains a minimum, False means it 
        contains a maximum.
    f0 : ndarray
        f(jd0). Providing this saves an extra function call
    f1 : ndarray
        f(jd1). Providing this saves an extra function call
    tol : float
        Tolerance used to determine when convergence is complete.
        
    Returns
    -------
    jd : ndarray
        the jd values at which f(jd) is a maximum or minimum.
    """
    if isinstance(minimum, bool):
        if minimum:
            multiplier = numpy.ones_like(jd0)
        else:
            multiplier = -numpy.ones_like(jd0)
    else:
        multiplier = numpy.ones_like(minimum, dtype=int)
        multiplier[~minimum] = -1
    
    maxiter = 500
    mintol = 1e-11
    
    def g(x, multiplier):
        return multiplier*f(x) + (multiplier==1)*360
    
    xa, xb, xc, fa, fb, fc = _bracket(f, jd0, jd1, multiplier, f0, f1)
    _cg = 0.3819660
    x = numpy.copy(xb)
    w = numpy.copy(xb)
    v = numpy.copy(xb)
    fw = numpy.copy(fb)
    fv = numpy.copy(fb)
    fx = numpy.copy(fb)
    
    a = numpy.zeros_like(xa)
    b = numpy.zeros_like(xb)
    case1 = xa < xc
    case2 = ~ case1
    a[case1] = xa[case1]
    b[case1] = xc[case1]
    a[case2] = xc[case2]
    b[case2] = xa[case2]

    deltax =  numpy.zeros_like(a)
    u =  numpy.zeros_like(a)
    fu = numpy.zeros_like(a)
    rat = numpy.zeros_like(a)
    tmp1 = numpy.zeros_like(a)
    tmp2 = numpy.zeros_like(a)
    p = numpy.zeros_like(a)
    dx_temp = numpy.zeros_like(a)

    iter_ = 0
    while (iter_ < maxiter):
        tol1 = tol * numpy.abs(x) + mintol
        tol2 = 2.0 * tol1
        xmid = 0.5 * (a + b)
        
        converged = numpy.abs(x - xmid) < (tol2 - 0.5 * (b - a))
        if converged.all():
            break
        not_converged = numpy.nonzero(~converged)[0]

        for i in not_converged:
            if (numpy.abs(deltax[i]) <= tol1[i]):
                if (x[i] >= xmid[i]):
                    deltax[i] = a[i] - x[i]       # do a golden section step
                else:
                    deltax[i] = b[i] - x[i]
                rat[i] = _cg * deltax[i]
            else:                              # do a parabolic step
                tmp1[i] = (x[i] - w[i]) * (fx[i] - fv[i])
                tmp2[i] = (x[i] - v[i]) * (fx[i] - fw[i])
                p[i] = (x[i] - v[i]) * tmp2[i] - (x[i] - w[i]) * tmp1[i]
                tmp2[i] = 2.0 * (tmp2[i] - tmp1[i])
                if (tmp2[i] > 0.0):
                    p[i] = -p[i]
                tmp2[i] = numpy.abs(tmp2[i])
                dx_temp[i] = deltax[i]
                deltax[i] = rat[i]
                # check parabolic fit
                if ((p[i] > tmp2[i] * (a[i] - x[i])) and (p[i] < tmp2[i] * (b[i] - x[i])) and
                        (numpy.abs(p[i]) < numpy.abs(0.5 * tmp2[i] * dx_temp[i]))):
                    rat[i] = p[i] * 1.0 / tmp2[i]        # if parabolic step is useful.
                    u[i] = x[i] + rat[i]
                    if ((u[i] - a[i]) < tol2[i] or (b[i] - u[i]) < tol2[i]):
                        if xmid[i] - x[i] >= 0:
                            rat[i] = tol1[i]
                        else:
                            rat[i] = -tol1[i]
                else:
                    if (x[i] >= xmid[i]):
                        deltax[i] = a[i] - x[i]  # if it's not do a golden section step
                    else:
                        deltax[i] = b[i] - x[i]
                    rat[i] = _cg * deltax[i]
    
            if (numpy.abs(rat[i]) < tol1[i]):            # update by at least tol1
                if rat[i] >= 0:
                    u[i] = x[i] + tol1[i]
                else:
                    u[i] = x[i] - tol1[i]
            else:
                u[i] = x[i] + rat[i]
                
        fu[not_converged] = g(u[not_converged], multiplier[not_converged])      # calculate new output value

        for i in not_converged:
            if (fu[i] > fx[i]):                 # if it's bigger than current
                if (u[i] < x[i]):
                    a[i] = u[i]
                else:
                    b[i] = u[i]
                if (fu[i] <= fw[i]) or (w[i] == x[i]):
                    v[i] = w[i]
                    w[i] = u[i]
                    fv[i] = fw[i]
                    fw[i] = fu[i]
                elif (fu[i] <= fv[i]) or (v[i] == x[i]) or (v[i] == w[i]):
                    v[i] = u[i]
                    fv[i] = fu[i]
            else:
                if (u[i] >= x[i]):
                    a[i] = x[i]
                else:
                    b[i] = x[i]
                v[i] = w[i]
                w[i] = x[i]
                x[i] = u[i]
                fv[i] = fw[i]
                fw[i] = fx[i]
                fx[i] = fu[i]

        iter_ += 1
    return x



from skyfield import almanac
from skyfield import api

ts = api.load.timescale()
t0 = ts.tt(-1000, 1, 1)
t1 = ts.tt(2000, 1, 1)
days = int(t1 - t0)

if 1:
    t = ts.tt(-1000, 1, range(days))
    gy, gm, gd = t.tt_calendar()[:3]
    ts.julian_calendar_cutoff = 99999999999999999 #api.GREGORIAN_START
    jy, jm, jd = t.tt_calendar()[:3]
    print(gd - jd)

    # import numpy as np
    # t = np.arange(0.0, 2.0, 0.01)
    # s = 1 + np.sin(2 * np.pi * t)

    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()
    ax.plot(t.J, gd - jd, '.') #, label='label', linestyle='--')
    ax.axvline(325, color='blue')
    #ax.plot(325, 0, '.') #, label='label', linestyle='--')
    # ax.set(xlabel='time (s)', ylabel='voltage (mV)', title='Title')
    # ax.set_aspect(aspect=1.0)
    ax.grid()
    # plt.legend()
    ax.set_ylim(-10, 15)
    fig.savefig('tmp.png')
    exit()

eph = api.load('de422.bsp')
t, y = almanac.find_discrete(t0, t1, almanac.seasons(eph))

vernal = (y == 0)
t = t[vernal]
y = y[vernal]

#cal = t.ut1_calendar
cal = lambda: t.utc
gday = cal()[2]
for yi, ti in zip(y, t):
    if 325 < ti.J <= 326:
        print(yi, almanac.SEASON_EVENTS[yi], ti.utc_iso(' '))

ts.julian_calendar_cutoff = api.GREGORIAN_START

del t.utc
jday = cal()[2]
for yi, ti in zip(y, t):
    if 325 < ti.J <= 326:
        print(yi, almanac.SEASON_EVENTS[yi], ti.utc_iso(' '))

import matplotlib.pyplot as plt
fig, ax = plt.subplots()
ax.plot(t.J, jday, 'o')
ax.plot(t.J, gday, '.') #, alpha=0.25)
ax.grid()
ax.axvline(325, color='r')
fig.savefig('tmp.png')



"""Check Skyfield lunar eclipses against the huge NASA table of eclipses."""

from collections import Counter
from skyfield import eclipselib
from skyfield.api import load, GREGORIAN_START
import datetime as dt

with load.open('https://eclipse.gsfc.nasa.gov/5MCLE/5MKLEcatalog.txt') as f:
    data = f.read()

ts = load.timescale()
ts.julian_calendar_cutoff = GREGORIAN_START

table = []
lines = data.decode('ascii').splitlines()

for line in lines[14:]:
    year = int(line[7:12])
    #if year < 1 or year >= 3000:
    #if year < 1000 or year > 2500:  # Computes the claim made in `almanac.rst`.
    #if year < 1900 or year > 1920:
    if year >= 3000:  # Minimum necessary to work: DE422 stops at year 3000.
       continue
    datestr = line[13:29]
    d = dt.datetime.strptime('2024 ' + datestr, '%Y %b %d  %H:%M:%S')
    t = ts.tt(year, d.month, d.day, d.hour, d.minute, d.second)
    table.append((t, line[51]))

time0 = table[0][0]
timeN = table[-1][0]

start_time = ts.tt_jd(time0.whole - 1, time0.tt_fraction)
end_time = ts.tt_jd(timeN.whole + 1, timeN.tt_fraction)

print('Start time:', start_time.utc_iso(' '))
print('End time:  ', end_time.utc_iso(' '))

start_time.whole -= 1.0
end_time.whole += 1.0

eph = load('de422.bsp')
#print(eph)  # To double-check which dates it covers, in case of doubt.
t, y, details = eclipselib.lunar_eclipses(start_time, end_time, eph)

max_diff = 0.0
total_diff = 0.0
problems = Counter()

i = 0

for ti, yi in zip(t, y):
    letter2 = 'NPT'[yi]
    t5, letter = table[i]
    diff = (ti - t5) * 24.0 * 3600.0
    #diff = (ti.utc_datetime() - t5.utc_datetime()).total_seconds()
    if diff < -100.0:
        problems['Extra'] += 1
        print(ti.utc_iso(), yi, 'EXTRA:not in catalog', letter2)
        continue
    while diff > 100.0:
        problems['Missing'] += 1
        print(' ' * len(t5.utc_iso()), '-', t5.utc_iso(), letter,
              'MISSING eclipse!')
        i += 1
        t5, letter = table[i]
        diff = (ti - t5) * 24.0 * 3600.0
    max_diff = max(max_diff, abs(diff))
    total_diff += abs(diff)
    judge = '' if letter == letter2 else f'their {letter} != our {letter2}'
    if judge:
        problems['Mismatch'] += 1
    print(ti.utc_iso(), yi, t5.utc_iso(), letter, diff, judge)
    i += 1

print('Largest difference in time of an eclipse (seconds):', max_diff)
print('Total difference in seconds between eclipse times:', total_diff)
for name, count in sorted(problems.items()):
    print(f'{name}: {count} / {len(table)}')



#!/usr/bin/env python

from time import time

from skyfield.api import load
from skyfield.data import mpc

def main():
    f = load.open(mpc.COMET_URL)

    t0 = time()
    c = mpc.load_comets_dataframe(f)
    print(time() - t0, 'seconds for load_comets_dataframe()')
    assert len(c) == 864

    f.seek(0)

    t0 = time()
    c = mpc.load_comets_dataframe_slow(f)
    print(time() - t0, 'seconds for load_comets_dataframe_slow()')
    assert len(c) == 864

    # f = load.open(mpc.MPCORB_URL)

    # t0 = time()
    # mpc.load_mpcorb_dataframe(f, slow=False)
    # print(time() - t0, 'seconds for load_mpcorb_dataframe() with slow=False')

    # f.seek(0)

    # t0 = time()
    # mpc.load_mpcorb_dataframe(f, slow=)
    # print(time() - t0, 'seconds for load_mpcorb_dataframe() with slow=')

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Routines to download Earth orientation data."""

import numpy as np
import os
from skyfield.api import load
from skyfield.timelib import julian_date

def morrison_and_stephenson_2004_table():
    """Table of smoothed Delta T values from Morrison and Stephenson, 2004."""
    import pandas as pd
    f = load.open('http://eclipse.gsfc.nasa.gov/SEcat5/deltat.html')
    tables = pd.read_html(f.read())
    df = tables[0]
    return pd.DataFrame({'year': df[0], 'delta_t': df[1]})

def parse_S15_table(f):
    """Parse polynomial coefficients from Table S15.

    The table is available at the website of Her Majesty's Nautical
    Almanac Office, from the paper “Measurement of the Earth's Rotation:
    720 BC to AD 2015” by L.V. Morrison, F.R. Stephenson, C.Y. Hohenkerk
    and M. Zawilski 2021.

    """
    # http://astro.ukho.gov.uk/nao/lvm/Table-S15.2020.txt
    content = f.read()
    banner = b'- ' * 36 + b'-\n'
    sections = content.split(banner)
    names = sections[1].splitlines()[-1].decode('utf-8').split()
    table = np.loadtxt(sections[2].splitlines())
    return names, table.T

def main():
    thisdir = os.path.dirname(__file__)

    df = morrison_and_stephenson_2004_table()
    year = df.year.values
    jd = julian_date(year, 1, 1)
    delta_t = df.delta_t.values
    array = np.array((jd, delta_t))
    np.save(os.path.join(thisdir, 'morrison_stephenson_deltat.npy'),
            array)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import numpy as np
from skyfield.api import Time, load

ATTRIBUTES = (
    'J', 'delta_t', 'dut1', 'gmst',
    # (lambda t: t.toordinal()),
    'tai_fraction', 'tdb_fraction', 'ut1_fraction',
)

def main():
    ts = load.timescale()
    t = ts.utc(2020, 10, 24)
    for attribute in ATTRIBUTES:
        step_width, up, down = measure_step_width(ts, t.whole, attribute)
        step_width_s = step_width * 24 * 60 * 60
        print('{:14}  {:12.6g} seconds  {} steps up, {} steps down'.format(
            attribute, step_width_s, up, down))

def measure_step_width(ts, whole, attribute):
    samples = np.arange(-500, 501)
    widths = []

    for tenth in 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0:
        exp = -30
        while True:
            tt_fraction = tenth + samples * 10 ** exp
            t = Time(ts, whole, tt_fraction)
            output = getattr(t, attribute)
            diff = np.diff(output)
            steps_up = (diff > 0.0).sum()
            steps_down = (diff < 0.0).sum()
            steps = steps_up + steps_down
            if steps >= 10:
                break
            exp += 1
        step_width = (tt_fraction[-1] - tt_fraction[0]) / steps
        widths.append(step_width)

    return max(widths), steps_up, steps_down

if __name__ == '__main__':
    main()



"""How accurate is Skyfield's subpoint computation routine?

Let's call it with a varying number of iterations, and see how
accurately it can turn a Topos-generated position back into a latitude,
longitude, and elevation.

"""
from numpy import einsum
from skyfield.api import Topos, load
from skyfield.constants import AU_M, DEG2RAD
from skyfield.earthlib import reverse_terra

def main():
    ts = load.timescale()
    t = ts.tt(2018, 1, 22, 9, 9, 20)
    trial_angles = 10, 20, 30, 40  # the error peaks around 20 degrees
    trial_elevations = 0, 6000, AU_M
    print(__doc__)
    for n in range(1, 5):
        print('=== {} iterations ==='.format(n))
        print('')
        for elevation_m in trial_elevations:
            for degrees in trial_angles:
                top = Topos(latitude_degrees=degrees, longitude_degrees=123,
                            elevation_m=elevation_m)
                xyz_au = top.at(t).position.au
                xyz_au = einsum('ij...,j...->i...', t.M, xyz_au)
                lat, lon, elev = reverse_terra(xyz_au, t.gast, n)
                lat = lat / DEG2RAD
                error_mas = 60.0 * 60.0 * 1000.0 * abs(degrees - lat)
                print('latitude {} degrees, elevation {} m'
                      ' -> error of {:.2f} mas'
                      .format(degrees, elevation_m, error_mas))
        print('')
    print("""\
Given that iterations=3 pushes the maximum error from tens of mas ("mas"
means " milli-arcsecond") down to hundredths of a mas, it is the value
we have chosen as a default.  A fourth iteration, if we ever chose to
perform one, pushes the error down to "0.00 mas".
""")

if __name__ == '__main__':
    main()



# mpc_make_excerpt.py

"""Search the MPCORB file for minor planets, given their packed designations."""

import argparse
import re
import sys
import zlib

from skyfield.api import load
from skyfield.data import mpc

def main(argv):
    parser = argparse.ArgumentParser(description='Grep MPCORB.DAT.gz')
    parser.add_argument('designations', nargs='+', help='packed designations'
                        ' of the minor planets whose orbits you need')
    args = parser.parse_args(argv)

    designations = [re.escape(d.encode('ascii')) for d in args.designations]
    pattern = rb'^((?:%s) .*\n)' % rb'|'.join(designations)
    r = re.compile(pattern, re.M)

    data = load.open(mpc.MPCORB_URL).read()
    data = zlib.decompress(data, wbits = zlib.MAX_WBITS | 16)
    lines = r.findall(data)

    sys.stdout.buffer.write(b''.join(lines))

if __name__ == '__main__':
    main(sys.argv[1:])



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.dates import HourLocator, DateFormatter

from skyfield.api import load, EarthSatellite

# Labels for both date and hour on the x axis, and km on y.

def label_dates_and_hours(axes):
    axes.xaxis.set_major_locator(HourLocator([0]))
    axes.xaxis.set_minor_locator(HourLocator([0, 6, 12, 18]))
    axes.xaxis.set_major_formatter(DateFormatter('0h\n%Y %b %d\n%A'))
    axes.xaxis.set_minor_formatter(DateFormatter('%Hh'))
    for label in ax.xaxis.get_ticklabels(which='both'):
        label.set_horizontalalignment('left')
    axes.yaxis.set_major_formatter('{x:.0f} km')
    axes.tick_params(which='both', length=0)

# Load the satellite's final TLE entry.

sat = EarthSatellite(
    '1 34602U 09013A   13314.96046236  .14220718  20669-5  50412-4 0   930',
    '2 34602 096.5717 344.5256 0009826 296.2811 064.0942 16.58673376272979',
    'GOCE',
)

# Build the time range `t` over which to plot, plus other values.

ts = load.timescale()
t = ts.tt_jd(np.arange(sat.epoch.tt - 2.0, sat.epoch.tt + 2.0, 0.005))
reentry = ts.utc(2013, 11, 11, 0, 16)
earth_radius_km = 6371.0

# Compute geocentric positions for the satellite.

g = sat.at(t)
valid = [m is None for m in g.message]

# Start a new figure.

fig, ax = plt.subplots()

# Draw the blue curve.

x = t.utc_datetime()
y = np.where(valid, g.distance().km - earth_radius_km, np.nan)
ax.plot(x, y)

# Label the TLE epoch.

x = sat.epoch.utc_datetime()
y = sat.at(sat.epoch).distance().km - earth_radius_km
ax.plot(x, y, 'k.')
ax.text(x, y - 9, 'Epoch of TLE data ', ha='right')

# Label the official moment of reentry.

x = reentry.utc_datetime()
y = sat.at(reentry).distance().km - earth_radius_km
ax.plot(x, y, 'r.')
ax.text(x, y + 6, ' Moment of re-entry', c='r')

# Grid lines and labels.

ax.grid(which='both')
ax.set(title='GOCE satellite: altitude above sea level', xlabel='UTC')
label_dates_and_hours(ax)

# Render the plot to a PNG file.

fig.savefig('goce-reentry.png', bbox_inches='tight')



import math
import numpy as np
from numpy import around
from time import time

print('Which approach computes "units:minutes:sections.fraction" faster?')

places = 4
#places = 0
#places = 1

# Original approach: integers with %.

def a(value):
    t0 = time()
    power = 10 ** places
    n = int(7200.0 * power * value + 1.0) // 2
    sign = n
    n = abs(n)
    n, fraction = divmod(n, power)
    n, seconds = divmod(n, 60)

    n, minutes = divmod(n, 60)
    sign = '-' if sign < 0 else ''
    s = '%s%02dh %02dm %02d.%0*ds' % (
        sign, n, minutes, seconds, places, fraction)
    dt = time() - t0
    return dt, s

# Variant of original, just to check difference: integers with "format".

def b(value):
    t0 = time()
    power = 10 ** places
    n = m = int(7200.0 * power * value + 1.0) // 2
    n = abs(n)
    n, fraction = divmod(n, power)
    n, seconds = divmod(n, 60)

    n, minutes = divmod(n, 60)
    sign = '-' if m < 0 else ''
    s = '{0}{1:02}h {2:02}m {3:02}.{4:0{5}}s'.format(
        sign, n, minutes, seconds, fraction, places)
    dt = time() - t0
    return dt, s

# Possible new approach: floats, using round() to set the number of
# digits.  (Wow, turns out to be more expensive than I had imagined!
# The round() call is by itself the cost of either of the other
# solutions in their entirety.)

def c(value):
    t0 = time()
    value = value * 3600.0
    value = round(value, places)
    minutes, seconds = divmod(abs(value), 60.0)
    minutes = int(minutes)

    units, minutes = divmod(minutes, 60)
    sign = '-' if value < 0 else ''
    s = '{0}{1:02}h {2:02}m {3:0{4}.{5}f}s'.format(
        sign, units, minutes, seconds, places + 3, places)
    dt = time() - t0
    return dt, s

for value in 1, -1, math.pi, -0.000000001:
    value = np.array(value)
    for approach in a, b, c:
        duration, s = approach(value)
        print('{:10.8f}  {}'.format(duration, s))
    print()

print('Comparison of early int() vs late int():')

# Wow, even though the early int() requires an extra abs() operation on
# the floating point `value`, it's still faster overall.

value = math.pi
power = 10000

n = int(power * 7200 * abs(value) + 1.0) // 2
t0 = time()
n = int(power * 7200 * abs(value) + 1.0) // 2
print(time() - t0, 'early')

n = int((power * 7200 * value + 1.0) // 2.0)
t0 = time()
n = int((power * 7200 * value + 1.0) // 2.0)
print(time() - t0, 'late')

n = int((power * 3600 * value + 0.5) // 1.0)
t0 = time()
n = int((power * 3600 * value + 0.5) // 1.0)
print(time() - t0, 'late 2')

n = int((power * 7200 * value + 1.0) // 2.0)
t0 = time()
n = int((power * 7200 * value + 1.0) // 2.0)
print(time() - t0, 'late 3')



from skyfield.api import load, Topos, EarthSatellite, Star
from almanac2 import seasons, moon_phases, meridian_transits, culminations, twilights, risings_settings

ts = load.timescale()

ephem = load('de430t.bsp')
earth = ephem['earth']
sun = ephem['sun']
moon = ephem['moon']
mars = ephem['mars barycenter']

greenwich = earth + Topos(latitude_degrees=(51, 28, 40), longitude_degrees=(0, 0, -5))

iss_tle = """\
1 25544U 98067A   18161.85073725  .00003008  00000-0  52601-4 0  9993
2 25544  51.6418  50.3007 0003338 171.6979 280.7366 15.54163173117534
"""
ISS = EarthSatellite(*iss_tle.splitlines())

sirius = Star(ra_hours=(6, 45, 8.91728), dec_degrees=(-16, 42, 58.0171))

t0 = ts.utc(2017, 1, 1)
t1 = ts.utc(2017, 1, 8)


# Equinoxes and Solstices
season_times, lons = seasons(earth, ts.utc(2000), ts.utc(2026))

# Moon Phases
phase_times, lon_diffs = moon_phases(moon, ts.utc(2018, 1), ts.utc(2018, 2))

# Rise and Set Times
rise_set_times, rise_or_set = risings_settings(greenwich, sun, t0, t1)

# Culminations
culmination_times, kinds = culminations(greenwich, sun, t0, t1)

# Meridian Transits
transit_times, hour_angles = meridian_transits(greenwich, mars, t0, t1)

# Twilights
naut_twilight_times, am_pm = twilights(greenwich, sun, t0, t1, kind='nautical')

# You can submit different object types to the same function:
for name, body in zip(['Sun', 'Mars', 'Sirius', 'ISS'], [sun, mars, sirius, ISS]):
    times, kinds = risings_settings(greenwich, body, t0, t1)
    print (name, 'first rises at:', times[kinds=='rise'][0].utc_jpl())



"""Basic operations that are needed repeatedly throughout Skyfield."""

from numpy import (
    arcsin, arctan2, array, cos, einsum, finfo, float64,
    full_like, load, nan, rollaxis, sin, sqrt, where,
)
from pkgutil import get_data
from skyfield.constants import tau

_AVOID_DIVIDE_BY_ZERO = finfo(float64).tiny

class A(object):
    """Allow literal NumPy arrays to be spelled ``A[1, 2, 3]``."""
    __getitem__ = array
A = A()

def sqrt_nan(n):
    """Return the square root of ``n``, or ``nan`` if ``n < 0``."""
    # (See design/sqrt_nan.py for a speed comparison of approaches.)
    return where(n < 0.0, nan, sqrt(abs(n)))

def dots(v, u):
    """Given one or more vectors in `v` and `u`, return their dot products.

    This works whether `v` and `u` each have the shape ``(3,)``, or
    whether they are each whole arrays of corresponding x, y, and z
    coordinates and have shape ``(3, N)``.

    """
    return (v * u).sum(axis=0)

def T(M):
    """Swap the first two dimensions of an array."""
    return rollaxis(M, 1)

def mxv(M, v):
    """Matrix times vector: multiply an NxN matrix by a vector."""
    return einsum('ij...,j...->i...', M, v)

def mxm(M1, M2):
    """Matrix times matrix: multiply two NxN matrices."""
    return einsum('ij...,jk...->ik...', M1, M2)

def mxmxm(M1, M2, M3):
    """Matrix times matrix times matrix: multiply 3 NxN matrices together."""
    return einsum('ij...,jk...,kl...->il...', M1, M2, M3)

_T, _mxv, _mxm, _mxmxm = T, mxv, mxm, mxmxm  # In case anyone imported old name

def length_of(xyz):
    """Given a 3-element array |xyz|, return its length.

    The three elements can be simple scalars, or the array can be two
    dimensions and offer three whole series of x, y, and z coordinates.

    """
    return sqrt((xyz * xyz).sum(axis=0))

def angle_between(u, v):
    """Given two vectors `v` and `u`, return the radian angle separating them.

    This works whether `v` and `u` each have the shape ``(3,)``, or
    whether they are each whole arrays of corresponding x, y, and z
    coordinates with shape ``(3, N)``. The returned angle will be
    between 0 and tau/2.

    This formula is from Section 12 of:
    https://people.eecs.berkeley.edu/~wkahan/Mindless.pdf

    """
    a = u * length_of(v)
    b = v * length_of(u)
    return 2.0 * arctan2(length_of(a - b), length_of(a + b))

def to_spherical(xyz):
    """Convert |xyz| to spherical coordinates (r,theta,phi).

    ``r`` - vector length
    ``theta`` - angle above (+) or below (-) the xy-plane
    ``phi`` - angle around the z-axis

    Note that ``theta`` is an elevation angle measured up and down from
    the xy-plane, not a polar angle measured from the z-axis, to match
    the convention for both latitude and declination.

    """
    r = length_of(xyz)
    x, y, z = xyz
    theta = arcsin(z / (r + _AVOID_DIVIDE_BY_ZERO))
    phi = arctan2(y, x) % tau
    return r, theta, phi

def _to_spherical_and_rates(r, v):
    # Convert Cartesian rate and velocity vectors to angles and rates.
    x, y, z = r
    xdot, ydot, zdot = v

    length = length_of(r)
    lat = arcsin(z / (length + _AVOID_DIVIDE_BY_ZERO))
    lon = arctan2(y, x) % tau
    range_rate = dots(r, v) / length_of(r)

    x2 = x * x
    y2 = y * y
    x2_plus_y2 = x2 + y2 + _AVOID_DIVIDE_BY_ZERO
    lat_rate = (x2_plus_y2 * zdot - z * (x * xdot + y * ydot)) / (
        (x2_plus_y2 + z*z) * sqrt(x2_plus_y2))
    lon_rate = (x * ydot - xdot * y) / x2_plus_y2

    return length, lat, lon, range_rate, lat_rate, lon_rate

def from_spherical(r, theta, phi):
    """Convert (r,theta,phi) to Cartesian coordinates |xyz|.

    ``r`` - vector length
    ``theta`` - angle in radians above (+) or below (-) the xy-plane
    ``phi`` - angle in radians around the z-axis

    Note that ``theta`` is an elevation angle measured up and down from
    the xy-plane, not a polar angle measured from the z-axis, to match
    the convention for both latitude and declination.

    """
    rxy = r * cos(theta)
    return array((rxy * cos(phi), rxy * sin(phi), r * sin(theta)))

# Support users who might have imported these under their old names.
# I'm not sure why I called what are clearly spherical coordinates "polar".
to_polar = to_spherical
from_polar = from_spherical

def rot_x(theta):
    c = cos(theta)
    s = sin(theta)
    zero = theta * 0.0
    one = zero + 1.0
    return array(((one, zero, zero), (zero, c, -s), (zero, s, c)))

def rot_y(theta):
    c = cos(theta)
    s = sin(theta)
    zero = theta * 0.0
    one = zero + 1.0
    return array(((c, zero, s), (zero, one, zero), (-s, zero, c)))

def rot_z(theta):
    c = cos(theta)
    s = sin(theta)
    zero = theta * 0.0
    one = zero + 1.0
    return array(((c, -s, zero), (s, c, zero), (zero, zero, one)))

def angular_velocity_matrix(angular_velocity_vector):
    x, y, z = angular_velocity_vector
    zero = x * 0.0
    return array(((zero, -z, y), (z, zero, -x), (-y, x, zero)))

def _to_array(value):
    """Convert plain Python sequences into NumPy arrays.

    This lets users pass plain old Python lists and tuples to Skyfield,
    instead of always having to remember to build NumPy arrays.  We pass
    any kind of generic sequence to the NumPy ``array()`` constructor
    and wrap any other kind of value in a NumPy ``float64`` object.

    """
    if hasattr(value, 'shape'):
        return value
    elif hasattr(value, '__len__'):
        return array(value)
    else:
        return float64(value)

def _reconcile(a, b):
    """Coerce two NumPy generics-or-arrays to the same number of dimensions."""
    an = getattr(a, 'ndim', 0)
    bn = getattr(b, 'ndim', 0)
    difference = bn - an
    if difference > 0:
        if an:
            a.shape += (1,) * difference
        else:
            a = full_like(b, a)
    elif difference < 0:
        if bn:
            b.shape += (1,) * -difference
        else:
            b = full_like(a, b)
    return a, b

try:
    from io import BytesIO
except:
    from StringIO import StringIO as BytesIO

def load_bundled_npy(filename):
    """Load a binary NumPy array file that is bundled with Skyfield."""
    data = get_data('skyfield', 'data/{0}'.format(filename))
    return load(BytesIO(data))



#!/usr/bin/env python3

import numpy as np
from skyfield.api import load

def main():
    ts = load.timescale()
    for year in range(1950, 2051, 5):
        t1 = ts.utc(year)
        t2 = ts.tt_jd(np.nextafter(t1.tt, 1e99))
        d = (t2.whole - t1.whole + t2.tt_fraction - t1.tt_fraction)
        print(year, ':', d * (24 * 60 * 60), 'seconds')

if __name__ == '__main__':
    main()



"""Tests of whether units behave."""

from assay import assert_raises
from numpy import array, nan
from skyfield.units import (
    Angle, Distance, Velocity, UnpackingError, WrongUnitError,
)

try:
    from astropy import units as u
except ImportError:
    u = None

def needs_astropy(test):
    """Skip `test` if AstroPy is not available."""
    return None if (u is None) else test

def test_degree_rounding():
    tenth = 0.1 / 60.0 / 60.0  # of an arcsecond

    assert str(Angle(degrees=tenth * -600.75)) == '-00deg 01\' 00.1"'
    assert str(Angle(degrees=tenth * -600.25)) == '-00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * -599.75)) == '-00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * -599.25)) == '-00deg 00\' 59.9"'

    assert str(Angle(degrees=tenth * -1.75)) == '-00deg 00\' 00.2"'
    assert str(Angle(degrees=tenth * -1.25)) == '-00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * -0.75)) == '-00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * -0.25)) == '00deg 00\' 00.0"'

    assert str(Angle(degrees=0.0)) == '00deg 00\' 00.0"'

    assert str(Angle(degrees=tenth * 0.25)) == '00deg 00\' 00.0"'
    assert str(Angle(degrees=tenth * 0.75)) == '00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * 1.25)) == '00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * 1.75)) == '00deg 00\' 00.2"'

    assert str(Angle(degrees=tenth * 599.25)) == '00deg 00\' 59.9"'
    assert str(Angle(degrees=tenth * 599.75)) == '00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * 600.25)) == '00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * 600.75)) == '00deg 01\' 00.1"'

def test_angle_scalar_strs():
    assert str(Angle(degrees=array(91))) == '''91deg 00' 00.0"'''
    assert str(Angle(degrees=array(91), signed=True)) == '''+91deg 00' 00.0"'''
    assert str(Angle(hours=array(12))) == '''12h 00m 00.00s'''

def test_angle_array_strs():
    h = Angle(hours=array([0.5, nan, -13]))
    d = Angle(degrees=h._degrees)

    assert str(h) == '3 values from 00h 30m 00.00s to -13h 00m 00.00s'
    assert str(d) == '''3 values from 07deg 30' 00.0" to -195deg 00' 00.0"'''

    with assert_raises(WrongUnitError):
        h.dstr()
        d.hstr()

    assert h.hstr() == d.hstr(warn=False) == [
        '00h 30m 00.00s',
        'nan',
        '-13h 00m 00.00s',
    ]
    assert d.dstr() == h.dstr(warn=False) == [
        '07deg 30\' 00.0"',
        'nan',
        '-195deg 00\' 00.0"',
    ]

    empty = Angle(radians=[])
    assert str(empty) == 'Angle []'
    assert empty.hstr(warn=False) == []
    assert empty.dstr() == []

    assert h.hstr(format='{0} {1} {2} {3} {4} {5}', places=6) == [
        ' 0 30 0 0 6', 'nan', '- 13 0 0 0 6']
    assert d.dstr(format='{0} {1} {2} {3} {4} {5}', places=6) == [
        ' 7 30 0 0 6', 'nan', '- 195 0 0 0 6']

def test_angle_sexagesimal_args():
    assert str(Angle(degrees=(90,))) == '''90deg 00' 00.0"'''
    assert str(Angle(hours=(12,))) == '''12h 00m 00.00s'''

    assert str(Angle(degrees=(90, 15))) == '''90deg 15' 00.0"'''
    assert str(Angle(hours=(12, 30))) == '''12h 30m 00.00s'''

    assert str(Angle(degrees=(90, 15, 30))) == '''90deg 15' 30.0"'''
    assert str(Angle(hours=(12, 30, 15))) == '''12h 30m 15.00s'''

def test_arcminutes_and_arcseconds_and_mas():
    angle = Angle(degrees=1.0)
    assert angle.arcminutes() == 60
    assert angle.arcseconds() == 60 * 60
    assert angle.mas() == 60 * 60 * 1000

def test_distance_input_units():
    for d in (
        Distance(1.0),
        Distance(au=1.0),       # deprecated
        Distance(m=149597870700),
        Distance(km=149597870.700),
        Distance.au(1.0),       # modern
        Distance.m(149597870700),
        Distance.km(149597870.700),
    ):
        assert abs(d.au - 1.0) <= 0

def test_velocity_input_units():
    v1 = Velocity(au_per_d=2.0)
    v2 = Velocity(km_per_s=3462.9137)
    assert abs(v1.au_per_d - v2.au_per_d) < 1e-7

    v1 = Velocity.au_per_d(2.0)
    v2 = Velocity.km_per_s(3462.9137)
    assert abs(v1.au_per_d - v2.au_per_d) < 1e-7

def test_stringifying_vector_distance():
    a = array([1.23, 4.56])
    s = str(Distance(au=a))
    if '[1' in s:
        # Python 3.5, says Travis CI.  No idea.
        assert s == '[1.23 4.56] au'
    else:
        # Every other version of Python.
        assert s == '[ 1.23  4.56] au'

def test_helpful_exceptions():
    distance = Distance(1.234)
    expect = '''\
to use this Distance, ask for its value in a particular unit:

    distance.au
    distance.km
    distance.m'''

    with assert_raises(UnpackingError) as a:
        x, y, z = distance
    assert str(a.exception) == expect

    with assert_raises(UnpackingError) as a:
        distance[0]
    assert str(a.exception) == expect

    velocity = Velocity(1.234)
    expect = '''\
to use this Velocity, ask for its value in a particular unit:

    velocity.au_per_d
    velocity.km_per_s
    velocity.m_per_s'''

    with assert_raises(UnpackingError) as a:
        x, y, z = velocity
    assert str(a.exception) == expect

    with assert_raises(UnpackingError) as a:
        velocity[0]
    assert str(a.exception) == expect

    angle = Angle(radians=1.234)
    expect = '''\
to use this Angle, ask for its value in a particular unit:

    angle.degrees
    angle.hours
    angle.radians'''

    with assert_raises(UnpackingError) as a:
        x, y, z = angle
    assert str(a.exception) == expect

    with assert_raises(UnpackingError) as a:
        angle[0]
    assert str(a.exception) == expect

def test_constructors_accept_plain_lists():
    Distance(au=[1,2,3])
    Distance(km=[1,2,3])
    Distance(m=[1,2,3])
    Velocity(au_per_d=[1,2,3])
    Velocity(km_per_s=[1,2,3])

def test_converting_from_km_to_m():
    distance = Distance(km=1.234)
    assert abs(distance.m - 1234.0) < 1e-15

def test_converting_from_m_to_km():
    distance = Distance(m=1234.0)
    assert abs(distance.km - 1.234) < 1e-15

def test_deprecated_method_from_au():
    distance = Distance.from_au(1.25)
    assert distance.au == 1.25

@needs_astropy
def test_converting_distance_with_astropy():
    distance = Distance(au=1.234)
    value1 = distance.km
    value2 = distance.to(u.km)
    epsilon = 0.02         # definitions of AU seem to disagree slightly
    assert abs(value1 - value2.value) < epsilon

@needs_astropy
def test_converting_velocity_with_astropy():
    velocity = Velocity(au_per_d=1.234)
    value1 = velocity.km_per_s
    value2 = velocity.to(u.km / u.s)
    epsilon = 1e-6
    assert abs(value1 - value2.value) < epsilon



#!/usr/bin/env python3
"""
The SGP4 library has the ability to process an array of satellites across an
array of times. What could support for this look like in Skyfield, combining
optimized parallel calculation (in the C++ version of SGP4) with coordinate
conversions from TEME?
"""
import numpy as np
from sgp4.api import SatrecArray, SGP4_ERRORS, accelerated
from skyfield.api import load, utc, EarthSatellite, Topos
from skyfield.constants import ANGVEL, AU_KM, DAY_S, AU_M, tau
from skyfield.sgp4lib import TEME_to_ITRF, _zero_zero_minus_one, theta_GMST1982, _cross
from skyfield.positionlib import build_position
from skyfield.vectorlib import ObserverData
from skyfield.functions import mxv, rot_z

# Precompute values to be reused across all satellites for the same times for conversion from TEME
def precompute_for_TEME(jd_ut1):
    theta, theta_dot = theta_GMST1982(jd_ut1)
    angular_velocity = np.multiply.outer(_zero_zero_minus_one, theta_dot)
    R = rot_z(-theta)
    return angular_velocity, R

# Based on an earlier version of TEME_to_ITRF() and does not include xp, yp, or fraction_ut1
def TEME_to_ITRF_fast(jd_ut1, rTEME, vTEME, angular_velocity, R):
    if len(rTEME.shape) == 1:
        rPEF = (R).dot(rTEME)
        vPEF = (R).dot(vTEME) + _cross(angular_velocity, rPEF)
    else:
        rPEF = mxv(R, rTEME)
        vPEF = mxv(R, vTEME) + _cross(angular_velocity, rPEF)
    return rPEF, vPEF

def precompute_for_ITRF(t):
    spin = rot_z(t.gast / 24.0 * tau)
    return spin

# Based on an earlier version of ITRF_to_GCRS2() and does not include _high_accuracy for now
def ITRF_to_GCRS2_fast(t, rITRF, vITRF, spin):
    position = mxv(spin, rITRF)
    velocity = mxv(spin, vITRF)
    velocity[0] += DAY_S * ANGVEL * - position[1]
    velocity[1] += DAY_S * ANGVEL * position[0]
    position = mxv(t.MT, position)
    velocity = mxv(t.MT, velocity)
    return position, velocity

# Run SGP4 in parallel across n satellites x m times
def positions_for(satellites, earthLocation, times):
    sat_array = SatrecArray( [s.model for s in satellites] )
    jd = times._utc_float()
    e, r, v = sat_array.sgp4(jd, np.zeros_like(jd))
    # Cached computations
    _, loc_v_GCRS, loc_p_GCRS, _ = earthLocation._at(times)
    loc_altaz_rotation = earthLocation._altaz_rotation(times)
    no_errors = [None] * len(e[0])
    angular_velocity, R = precompute_for_TEME(times.ut1)
    spin = precompute_for_ITRF(times)
    # Unpack the TEME coordinates, convert to GCRS and subtract earthLocation
    for index in range(r.shape[0]):
        errors = e[index]
        messages = no_errors if not np.any(errors) else \
                    [SGP4_ERRORS[error] if error else None for error in errors ]
        # Adapted from _position_and_velocity_TEME_km(), ITRF_position_velocity_error()
        rTEME = np.divide(r[index].T, AU_KM)
        vTEME = np.divide(v[index].T, AU_KM / DAY_S)
        rITRF, vITRF = TEME_to_ITRF_fast(times.ut1, rTEME, vTEME, angular_velocity, R)
        sat_p_GCRS, sat_v_GCRS = ITRF_to_GCRS2_fast(times, rITRF, vITRF, spin)
        #sat_p_GCRS, sat_v_GCRS = ITRF_to_GCRS2(times, rITRF, vITRF)
        # Mirror VectorSum _at() for (EarthSatellite - Topos)
        sat_p_GCRS -= loc_p_GCRS
        sat_v_GCRS -= loc_v_GCRS
        # Mirror VectorFunction at()
        observer_data = ObserverData()
        observer_data.gcrs_position = loc_p_GCRS
        observer_data.altaz_rotation = loc_altaz_rotation
        observer_data.elevation_m = earthLocation.elevation.m
        target = -100000 - satellites[index].model.satnum
        position = build_position(sat_p_GCRS, sat_v_GCRS, times, earthLocation, target, observer_data)
        position.message = messages
        yield position

# ------------------------------------------------------------------------- #

ISS_TLE = """1 25544U 98067A   20187.34541214  .00002866  00000-0  59793-4 0  9991
2 25544  51.6454 258.5877 0002990  76.8215 328.5773 15.49181247234835
"""

KESTREL_TLE = """1 42982U 98067NE  20187.19650118  .00008987  00000-0  76592-4 0  9999
2 42982  51.6323 190.7341 0002835 141.7806 218.3393 15.70424148153865
"""

def main():
    if not accelerated:
        print("SGP4 library is not accelerated")
        return
    test_iss()
    test_two_sats()
    test_many_sats()

def test_iss():
    ts = load.timescale(builtin=True)
    L1, L2 = ISS_TLE.splitlines()
    iss = EarthSatellite(L1, L2)
    trange = range(0, 60, 10)
    times = ts.utc(2020, 7, 5, 12, 0, trange)
    print("positions_for():")
    satellites = [iss]
    city = Topos('44.0247 N', '88.5426 W')
    for geo1 in positions_for(satellites, city, times):
        print(geo1.position.km)
    #
    print("Compare to EarthSatellite at():")
    difference = iss - city
    geo2 = difference.at(times)
    #geo2 = iss.at(times)
    print(geo2.position.km)
    #
    assert np.allclose(geo1.position.km, geo2.position.km)
    one_m_per_hour = 1.0 * 24.0 / AU_M
    assert abs(geo1.velocity.au_per_d - geo2.velocity.au_per_d).max() < one_m_per_hour

def test_two_sats():
    print("test two sats")
    ts = load.timescale(builtin=True)
    L1, L2 = ISS_TLE.splitlines()
    iss = EarthSatellite(L1, L2)
    L1, L2 = KESTREL_TLE.splitlines()
    kestrel = EarthSatellite(L1, L2)
    #
    city = Topos('44.0247 N', '88.5426 W')
    trange = range(0, 300, 10)
    times = ts.utc(2020, 7, 5, 12, 0, trange)
    satellites = [kestrel, iss]
    positions = list(positions_for(satellites, city, times))
    kestrel1, iss1 = positions
    iss2 = (iss - city).at(times)
    kestrel2 = (kestrel - city).at(times)
    assert np.allclose(iss1.position.km, iss2.position.km)
    assert np.allclose(kestrel1.position.km, kestrel2.position.km)
    assert np.allclose(iss1.velocity.au_per_d, iss2.velocity.au_per_d)
    assert np.allclose(kestrel1.velocity.au_per_d, kestrel2.velocity.au_per_d)

def test_many_sats():
    print("test many sats")
    ts = load.timescale(builtin=True)
    active_sats = load.tle_file("https://celestrak.org/NORAD/elements/active.txt", reload=False)
    print(" active satellites:", len(active_sats), " x 300 time values")
    trange = range(0, 300, 1)
    times = ts.utc(2020, 8, 25, 12, 0, trange)
    city = Topos('44.0247 N', '88.5426 W')
    positions = list(positions_for(active_sats, city, times))
    assert len(active_sats) == len(positions)

if __name__ == "__main__":
    main()
